from typing import Union, List, Tuple
from datetime import datetime
from dateutil.parser import parse
from dateutil.rrule import rrule, MONTHLY, MO, TU, WE, TH, FR, SA, SU
from dateutil.relativedelta import relativedelta
import calendar

def getDaysBetween(start: Union[str, datetime], end: Union[str, datetime],
                   month: List[int] = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                   weekday: List[int] = [MO, TU, WE, TH, FR, SA, SU],
                   weekno: List[int] = [], strftime: str = "%Y-%m-%d") -> List[Union[str, datetime]]:
    """Find a filtered date that exists within a period.

    Args:
        start (Union[str, datetime]): start day of a period
        end (Union[str, datetime]): end day of a period
        month (list[int], optional): month filter. Defaults to [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12].
        weekday (list[int], optional): week filter. An int type is also possible(0:monday ... 6:sunday). Defaults to [MO, TU, WE, TH, FR, SA, SU].
        weekno (list[int], optional): It means the sequence of weeks in which month. Defaults to [].
        strftime (str, optional): if set to None, returns datetime type. Defaults to "%Y-%m-%d".

    Returns:
        list[Union[str, datetime]]: Returns all dates in the period as a list in 'str' or 'datetime' format.
    """
    len_weekday = 1+len(weekday)
    result = rrule(
        MONTHLY,
        bymonth=month,
        byweekday=weekday,
        bysetpos=sum([list(range(len_weekday*(no-1), len_weekday*no)) for no in weekno], []),
        dtstart=parse(start) if isinstance(start, str) else start,
        until=parse(end) if isinstance(start, str) else end
    )
    
    return [x.strftime(strftime) if strftime else x for x in result]


def getDayPeriod(base: Union[str, datetime] = None,
                 off_year: int = 0, off_month: int = 0, off_day: int = 0, off_week: int = 0,
                 off_hour: int = 0, off_minute: int = 0, off_second: int = 0,
                 delta_year: int = 0, delta_month: int = 0, delta_day: int = 0, delta_week: int = 0,
                 delta_hour: int = 0, delta_minute: int = 0, delta_second: int = 0,
                 weekday: int = None, format: str = "%Y-%m-%d", as_str: bool = True) -> Tuple[Union[str, datetime]]:
    """Find the start date and end date that satisfy the condition.

    Args:
        base (Union[str, datetime], optional): The base date. if it is Null then means today. Defaults to None.
        off_year (int, optional): Change the base date by year. A negative number means before year. Defaults to 0.
        off_month (int, optional): Change the base date by month. A negative number means before month. Defaults to 0.
        off_day (int, optional): Change the base date by day. A negative number means before day. Defaults to 0.
        off_week (int, optional): Change the base date by week. A negative number means before week. Defaults to 0.
        delta_year (int, optional): The period in years based on the base date. Defaults to 0.
        delta_month (int, optional): The period in months based on the base date.. Defaults to 0.
        delta_day (int, optional): The period in days based on the base date.. Defaults to 0.
        delta_week (int, optional): The period in weeks based on the base date.. Defaults to 0.
        weekday (int, optional): Change the base date by a day of the week. Defaults to None.
        format (str): if set to None, returns datetime type. Defaults to "%Y-%m-%d".

    Returns:
        tuple[Union[str, datetime]]: Returns a tuple of period day in 'str' or 'datetime' format. (start day, end day)
    """
    if base is None:
        base_day = datetime.now()
    else:
        base_day = datetime.strptime(base, format) if isinstance(base, str) else base
    
    if weekday != None:
        weeks = [MO, TU, WE, TH, FR, SA, SU]
        if weekday in weeks:
            base_day = base_day + relativedelta(days=weeks.index(weekday)-base_day.weekday())
        else:
            base_day = base_day + relativedelta(days=weekday-base_day.weekday())

    start_day = base_day + relativedelta(years=off_year, months=off_month, days=off_day, weeks=off_week, hours=off_hour, minutes=off_minute, seconds=off_second)
    end_day = start_day + relativedelta(years=delta_year, months=delta_month, days=delta_day, weeks=delta_week, hours=delta_hour, minutes=delta_minute, seconds=delta_second)
    
    if start_day > end_day:
        temp = start_day
        start_day = end_day
        end_day = temp
    
    if as_str:
        start_day = start_day.strftime(format)
        end_day = end_day.strftime(format)
    
    return start_day, end_day


week_str = ['월', '화', '수', '목', '금', '토', '일']
def getWeekday(day: Union[str, datetime], as_str: bool = False):
    if isinstance(day, str):
        day = datetime.strptime(day, "%Y-%m-%d")
    weekday = day.weekday()
    return week_str[weekday] if as_str else weekday


animals=['원숭이','닭','개','돼지','쥐','소','범','토끼','용','뱀','말','양']
def getYearAnimal(year: Union[str, int]):
    if isinstance(year, str):
        year = int(year)
    return animals[year%12]

def getMonthStartEnd(year, month):
    assert month > 0 or month < 13
    
    start_date = f"{year}-{month:02d}-01"
    
    last_day = calendar.monthrange(year, month)[1]
    end_date = f"{year}-{month:02d}-{last_day}"
    
    return start_date, end_date


from psycopg2 import pool, extensions
from .IConnectionManager import IConnectionManager

class ConnMngPsql(IConnectionManager):
    def connect(self, **kwargs) -> IConnectionManager:
        try:
            self.close()
            self._kwargs.update(kwargs)
            
            self._conn_pool = pool.ThreadedConnectionPool(**self._kwargs)
            return self.instance
        except Exception:
            raise
    
    def isConnect(self) -> bool:
        return self._conn_pool is not None
    
    def close(self, **kwargs) -> None:
        if self._conn_pool is not None:
            self._conn_pool.closeall()
    
    def getConnection(self, auto_commit=True, **kwargs) -> extensions.connection:
        try:
            conn = self._conn_pool.getconn()
            conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT if auto_commit else extensions.ISOLATION_LEVEL_DEFAULT)
            return conn
        except Exception:
            raise
    
    def releaseConnection(self, conn) -> None:
        try:
            self._conn_pool.putconn(conn)
        except Exception:
            raise



import warnings, logging, os
import ipywidgets as ipw
import math

def asClassMethod(clazz):
    def makeMethod(func):
        setattr(clazz, func.__name__, func)
        if hasattr(clazz, '__abstractmethods__'):
            clazz.__abstractmethods__ = frozenset(
                name for name in clazz.__abstractmethods__
                if name != func.__name__
            )
        return func
    return makeMethod

def setWarningOff():
    warnings.filterwarnings("ignore")
    logging.getLogger("tensorflow").setLevel(logging.FATAL)
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

def showAsCols(*args, colSize=None, width="100%", margin="3px"):
    """Display multiple output data lines and columns as desired.
    Args:
      args: multiple datas.
      colSize: column size.
      width: css style width parameter. default is 100%. you can use "px" unint.
      margin: css style margin parameter. default is 3px.
    Returns:
      None
    Example:
      >>> showAsCols("a", "b", "c", colSize=2)
      a      b
      c
    """

    # calculate column and row counts.
    lenArgs = len(args)
    if (colSize is not None) and (colSize > 0) and (lenArgs - colSize > 0):
        lenRow = math.ceil(lenArgs/colSize)
        lenCol = colSize
    else:
        lenRow = 1
        lenCol = lenArgs

    # create Output() of ipywidgets and to display the args to it.
    layout = ipw.Layout(width=width, grid_gap=margin)
    grid = ipw.GridspecLayout(lenRow, lenCol, layout=layout)
    outs = [ipw.Output() for _ in range(len(args))]
    for idx, out in enumerate(outs):
        with out:
            display(args[idx])

    # allocate Output() to grid
    idxOuts = 0
    for row in range(lenRow):
        for col in range(lenCol):
            grid[row, col] = outs[idxOuts] if idxOuts < lenArgs else ipw.Output()
            idxOuts += 1
    
    display(grid)



from .plot_chart import (
    setJupyterEnable,
    getDataSource, getFigure, setFigureStyle,
    setLine, setBar, setMarke, setAreaBox, setHVLines,
    plotTimeseries, plotSeries
)

from .show import (
    showBokeh, showPandas,
    showAsRows, showAsGrid,
)




import pandas as pd
import numpy as np
from scipy.signal import butter, filtfilt
from typing import Union, List, Tuple, Callable

@pd.api.extensions.register_series_accessor("bi")
class DanbiExtendSeries:
    def __init__(self, column):
        self._col = column
        self._len = len(column)

    def rollapply(self, window: int, func: Callable, test: int = 0, future: bool = False, val_nan=np.nan):
        results = []
        if test > 0:
            for idx, df in enumerate(self._col.rolling(window)):
                if len(df) < window:
                    continue
                if idx + 1 >= window + test:
                    break
                results.append(func(df))

            return results
        else:
            for df in self._col.rolling(window):
                if len(df) < window:
                    continue
                results.append(func(df))
            
            if len(results) > 0:
                if isinstance(results[0], tuple):
                    nan = [[val_nan for _ in range(len(results[0]))] for _ in range(window-1)]
                else:
                    nan = [val_nan for _ in range(window-1)]
            
            return results + nan if future else nan + results

    def sigma(self, sigma: int = 3, quantile=0.1, as_int: bool = False) -> Tuple[float, float, float, float, float, float]:
        data = self._col.values
        val_max = max(data)
        val_min = min(data)
        mean = np.nanmean(data)
        std = np.nanstd(data)
        threshold = sigma * std

        sigma_lower = mean - threshold
        sigma_upper = mean + threshold
        sigma_percent = np.sum((data >= sigma_lower) & (data <= sigma_upper)) / data.size * 100

        quantile_lower = self._col.quantile(quantile)
        quantile_upper = self._col.quantile(1 - quantile)
        quantile_percent = np.sum((data >= quantile_lower) & (data <= quantile_upper)) / data.size * 100

        if as_int:
            lower = int(lower)
            mean = int(mean)
            upper = int(upper)
            std = int(std)
            threshold = int(threshold)
            sigma_percent = int(sigma_percent)
            quantile_percent = int(quantile_percent)

        return {
            "max": val_max,
            "min": val_min,
            "mean": mean,
            "std": std,
            "threshold": threshold,
            "sigma_lower": sigma_lower,
            "sigma_upper": sigma_upper,
            "sigma_percent": sigma_percent,
            "quantile_lower": quantile_lower,
            "quantile_upper": quantile_upper,
            "quantile_percent": quantile_percent
        }
    
    def lpf(self, cutoff: float = 0.1, fs: float = 1.0, order=5):
        data = self._col.values

        nyquist = 0.5 * fs
        normal_cutoff = cutoff / nyquist
        b, a = butter(order, normal_cutoff, btype='low', analog=False)
        y = filtfilt(b, a, data)

        return y
    
    def forceDirection(self, period: int = 1, sigma: Union[int, float, dict] = 1, zero_rate: float = 0.0):
        diff = self._col.diff(period)
        if isinstance(sigma, (int, float)):
            sigma = diff.bi.sigma(sigma)
        direction = (diff / sigma["threshold"]).clip(-1, 1).fillna(0)

        zero_bound = zero_rate / 100
        direction = direction.mask((direction >= -zero_bound) & (direction <= zero_bound), 0.0)

        return direction

    def stateUpDn(self, window: int, up_rate: float = 1, dn_rate: float = 1, future: bool = False) -> List:
        if future:
            ups = self._col == self._col.rolling(window).min().shift(-window+1)
            dns = self._col == self._col.rolling(window).max().shift(-window+1)
        else:
            ups = self._col == self._col.rolling(window).min()
            dns = self._col == self._col.rolling(window).max()
        updn, term, is_up, is_change, s_val, e_val = [], [], False, False, 0, 0

        for up, dn, val in zip(ups, dns, self._col):
            if not np.isnan(val):
                e_val = val
            if is_change:
                updn += term
                term, is_change = [], False
                s_val = val
            term.append(is_up)

            if not is_up and up: # dn -> up 으로 변환
                is_up, is_change = True, True
                if s_val != 0 and s_val * dn_rate < val: # dn상태가 rate_dn 조건을 만족하지 않음.
                    term = list(np.logical_not(term))
            elif is_up and dn: # up -> dn 으로 변환
                is_up, is_change = False, True
                if s_val != 0 and s_val * up_rate > val: # up상태가 rate_up 조건을 만족하지 않음.
                    term = list(np.logical_not(term))

        if (is_up and s_val * up_rate > e_val) or (not is_up and s_val * dn_rate < e_val):
            term = list(np.logical_not(term))
        updn += term
        
        return np.array(updn)

    def updnFuture(self, window: int, rate_up: float = 0.1, rate_dn: float = 0.1, valid_change: int = 3) -> List:
        ups = self._col == self._col.rolling(window).min().shift(-window+1)
        dns = self._col == self._col.rolling(window).max().shift(-window+1)
        updn, term, is_up, is_change, s_val, e_val = [], [], False, False, 0, 0
        change_period = 1

        for up, dn, val in zip(ups, dns, self._col):
            if not np.isnan(val):
                e_val = val
            if is_change:
                updn += term
                term, is_change = [], False
                s_val = val
                change_period = 1
            else:
                change_period += 1
            if change_period >= valid_change:
                term.append(None)
            else:
                term.append(is_up)

            if not is_up and up: # dn -> up 으로 변환
                is_up, is_change = True, True
                if s_val != 0 and s_val * (1-rate_dn) < val: # dn상태가 rate_dn 조건을 만족하지 않음.
                    term = list(np.logical_not(term))
            elif is_up and dn: # up -> dn 으로 변환
                is_up, is_change = False, True
                if s_val != 0 and s_val * (1+rate_up) > val: # up상태가 rate_up 조건을 만족하지 않음.
                    term = list(np.logical_not(term))

        if (is_up and s_val * (1+rate_up) > e_val) or (not is_up and s_val * (1-rate_dn) < e_val):
            term = list(np.logical_not(term))
        updn += term
        
        return updn

    def updnCurrent(self, window: int, rate_up: float = 0.1, rate_dn: float = 0.1, fillna: bool = True) -> List:
        updn = self._col.pct_change(window)
        updn.mask(updn > rate_up , 1, inplace=True)
        updn.mask(updn < -rate_dn, -1, inplace=True)
        updn.mask((updn <= rate_up) & (updn >= -rate_dn) & (updn < 0), -0.5, inplace=True)
        updn.mask((updn <= rate_up) & (updn >= -rate_dn) & (updn >= 0), 0.5, inplace=True)
        
        return updn.fillna(0) if fillna else updn





import tensorflow as tf
from tensorflow.io import TFRecordWriter
import numpy as np
from typing import List, Tuple, Dict, Callable


def getWriter(file_path: str, is_zip: bool = True) -> tf.io.TFRecordWriter:
    if bool:
        return tf.io.TFRecordWriter(file_path, tf.io.TFRecordOptions(compression_type="GZIP"))
    else:
        return tf.io.TFRecordWriter(file_path)

def storeTFRecord(writer: Callable, data: Dict[str, np.array], is_zip: bool = True):
    # for data in zip(*datas.values()):
    #     feature = {}
    #     for idx, name in enumerate(datas.keys()):
    #         feature[name] = tf.train.Feature(bytes_list=tf.train.BytesList(value=[data[idx].tobytes()]))
    #     example = tf.train.Example(features=tf.train.Features(feature=feature))
    #     writer.write(example.SerializeToString())

    feature = {}
    for idx, name in enumerate(data.keys()):
        feature[name] = tf.train.Feature(bytes_list=tf.train.BytesList(value=[data[name].tobytes()]))
    example = tf.train.Example(features=tf.train.Features(feature=feature))
    writer.write(example.SerializeToString())

def getTFRecordDecoder(datas: Dict[str, tf.DType], shape_in: Dict[str, Tuple[int]], shape_out: Dict[str, Tuple[int]]) -> Callable:
    feature_shape = {}
    for name in datas.keys():
        feature_shape[name] = tf.io.FixedLenFeature([], tf.string)
        
    def tfrecordParseAndDecode(dataset):
        features = tf.io.parse_single_example(dataset, features=feature_shape)
        
        ins = {}
        for name, shape in shape_in.items():
            raw = tf.io.decode_raw(features[name], datas[name])
            ins[name] = tf.reshape(raw, shape=shape)
        outs = {}
        for name, shape in shape_out.items():
            raw = tf.io.decode_raw(features[name], datas[name])
            outs[name] = tf.reshape(raw, shape=shape)

        return ins, outs
    return tfrecordParseAndDecode

def restoreTFRecord(files: List[str], decoder: Callable, is_zip: bool = True) -> tf.data.TFRecordDataset:
    if is_zip:
        rds = tf.data.TFRecordDataset(files, "GZIP")
    else:
        rds = tf.data.TFRecordDataset(files)
    
    rds = rds.map(decoder)
    
    return rds

def selectTFRecordIndexData(tfrecord, name, indices, axis=-1):
    """
    특정 키의 텐서에서 지정한 인덱스만 선택하여 새로운 데이터셋을 생성합니다.
    
    Args:
        tfrecord (tf.data.Dataset): 입력 데이터셋.
        name (str): 선택할 텐서가 저장된 딕셔너리 내의 키 이름.
        indices (list or tuple): 선택할 인덱스의 리스트.
        axis (int): 인덱스를 선택할 차원.
    
    Returns:
        tf.data.Dataset: 선택된 인덱스로 수정된 새로운 데이터셋.
    """
    # 인덱스를 텐서로 변환하여 map 함수 내에서 사용 가능하게 함
    indices_tensor = tf.constant(indices, dtype=tf.int32)
    
    def map_fn(*elements):
        """
        데이터셋의 각 요소에 적용될 매핑 함수.
        요소는 여러 개의 딕셔너리로 구성된 튜플일 수 있음.
        """
        new_elements = []
        for element in elements:
            new_dict = {}
            for k, v in element.items():
                if k == name:
                    # 지정한 키가 존재하면, 해당 텐서에서 지정한 축과 인덱스를 사용하여 선택
                    selected = tf.gather(v, indices_tensor, axis=axis)
                    new_dict[k] = selected
                else:
                    # 다른 키는 그대로 유지
                    new_dict[k] = v
            new_elements.append(new_dict)
        return tuple(new_elements)
    
    return tfrecord.map(map_fn)


import pandas as pd
from danbi.extends import bibokeh as bibo

def showLearningHistory(history, width: int = 1500, height: int = 400, legend: str = "bottm_left"):
    ylist = []
    keys = history.history.keys()
    for key in keys:
        if not key.startswith("val_"):
            validation = "val_"+key
            if validation in keys:
                ylist.append([key, "val_"+key])
            else:
                ylist.append([key])
            
    bibo.showPandas(
        pd.DataFrame(history.history), "index", ylist, "line", 1500, 400, True, "x", [2, 1],
        {"legend_location": legend}
    )


import numpy as np
import random
import danbi as bi
import threading


class UniqueRateChecker:
    def __init__(self):
        self._df_map = None
        self._map = None
    
    def store(self) -> dict:
        return {
            "df_map": self._df_map,
        }
    
    def storeFile(self, file: str) -> dict:
        bi.storePickle(self.store(), file)

    def getBalanceMap(self):
        return self._df_map
    
    def restore(self, states):
        self._df_map = states["df_map"]
        self._map = self._df_map.iloc[:, :-2].values
    
    def restoreFile(self, file: str):
        stored_dict = bi.restorePickle(file)
        self.restore(stored_dict)

    def fit(self, df, cols):
        df_map = df.groupby(cols).size().reset_index(name='counts')
        df_map["rate"] = (df_map["counts"].max() / (df_map["counts"])).astype(int)
        self._df_map = df_map.sort_values(by='counts', ascending=False).reset_index(drop=True)
        self._map = self._df_map.iloc[:, :-2].values

    def check(self, target):
        matches = np.ones(len(self._map), dtype=bool)
        for i in range(len(target)):
            matches &= (self._map[:, i] == target[i])
        indices = np.where(matches)[0]
        if len(indices) > 0:
            return self._df_map.iloc[indices[0]][["counts", "rate"]].astype(int).tolist()
        else:
            return [0, 0]

    def checkCounts(self, target):
        return self.check(target)[0]

    def checkRate(self, target):
        return self.check(target)[1]


class LabelRateAugmenter(UniqueRateChecker):
    def __init__(self, augmentation_rate: float = 1, noise_fraction: float = 0.1, noise_std: float = 0.1, noise_seed: int = None):
        self._aug_rate = augmentation_rate
        self._noise_fraction = noise_fraction
        self._noise_std = noise_std
        self._noise_rng = np.random.default_rng(noise_seed)
        self._noise_last_dim = None
        self._noise_size = None
        self._origins = []
        self._size = 0
        self._lock = threading.Lock()

    def store(self) -> dict:
        data = super().store()
        data["aug_rate"] = self._aug_rate
        data["noise_rate"] = self._noise_rate
        
        return data

    def restore(self, states):
        data = super().restore(states)
        self._aug_rate = states["aug_rate"]

    def add(self, data, label):
        if self._noise_last_dim is None:
            self._noise_last_dim = data.shape[-1]
            self._noise_size = int(self._noise_last_dim * self._noise_fraction)
        
        rate = self.checkRate(label) - 1
        rate = int(rate * self._aug_rate)
        self._size += rate
        
        if rate > 0:
            self._origins.append([data, label, rate])

    def getAddedDatas(self):
        return self._origins

    def _get_random(self):
        index = random.randint(0, len(self._origins) - 1)
        item = self._origins[index]
        
        if item[-1] == 1:
            del self._origins[index]
        else:
            self._origins[index][-1] -= 1
        
        return item[0], item[1]

    def __iter__(self):
        return self

    def __next__(self):
        with self._lock:
            if len(self._origins) == 0:
                raise StopIteration
            else:
                data, label = self._get_random()
                if self._noise_fraction is not None:
                    it = np.nditer(data[..., 0], flags=['multi_index'])
                    while not it.finished:
                        idx = it.multi_index # 현재 위치에서 마지막 차원에 대한 슬라이스 가져오기
                        noise_indices = self._noise_rng.choice(self._noise_last_dim, size=self._noise_size, replace=False) # 마지막 차원에서 랜덤하게 50% 인덱스를 선택
                        data[idx + (noise_indices,)] += self._noise_rng.normal(0, self._noise_std, size=self._noise_size) # 선택된 인덱스에 대해 노이즈 추가
                        
                        it.iternext()
                return data, label
    
    def __len__(self):
        return self._size



import abc
from typing import Any

class IMapper(abc.ABC):

    @abc.abstractclassmethod
    def setConfigPaths(self, conf_paths: list = None):
        ...
    
    @abc.abstractclassmethod
    def getConfigPaths(self) -> list:
        ...
    
    @abc.abstractclassmethod
    def setNamespaceTag(self, namespace: str, tag: str, base_package: str = None) -> None:
        ...
    
    @abc.abstractclassmethod
    def get(self, name: str, values: Any = None) -> str:
        ...




from typing import Union
import pandas as pd
from .IDB import IDB

class DBPsqlAsync(IDB):
    async def query(self, mapper_name: str, values: dict = None) -> list:
        with self._lock:
            raw_sql = self._mapper.get(mapper_name, values)
            if values is not None:
                raw_sql, values = self._pyformat2psql(raw_sql, values)
            
            return await self.queryRaw(raw_sql, values)
    
    async def queryRaw(self, raw_sql: str, values: list = None) -> list:
        try:
            with self._lock:
                conn = await self._manager.getConnection()
                if values is None:
                    records = await conn.fetch(raw_sql)
                else:
                    records = await conn.fetch(raw_sql, *values)
                await self._manager.releaseConnection(conn)
                
                return records
        except Exception:
            if conn is not None:
                await self._manager.releaseConnection(conn)
            raise
    
    async def queryPandas(self, mapper_name: str, values: Union[dict, tuple] = None, dtype: dict = None) -> pd.DataFrame:
        with self._lock:
            raw_sql = self._mapper.get(mapper_name, values)
            if values is not None:
                raw_sql, values = self._pyformat2psql(raw_sql, values)
            
            return await self.queryPandasRaw(raw_sql, values, dtype)
    
    async def queryPandasRaw(self, raw_sql: str, values: list = None, dtype: dict = None) -> pd.DataFrame:
        try:
            with self._lock:
                records = await self.queryRaw(raw_sql, values)
                if len(records) > 0:
                    df = pd.DataFrame(records, columns=records[0].keys())
                else:
                    return pd.DataFrame()
                
                return df if dtype is None else df.astype(dtype)
        except Exception:
            raise
    
    async def execute(self, mapper_name, values=None) -> None:
        with self._lock:
            raw_sql = self._mapper.get(mapper_name, values)
            await self.executeRaw(raw_sql, values)
    
    async def executeRaw(self, raw_sql, values=None) -> None:
        try:
            with self._lock:
                conn = await self._manager.getConnection()
                if values is None:
                    await conn.execute(raw_sql)
                else:
                    await conn.execute(raw_sql, *values)
                await self._manager.releaseConnection(conn)
        except Exception:
            if conn is not None:
                await self._manager.releaseConnection(conn)
            raise
    
    async def executeMany(self, mapper_name, values=None) -> None:
        with self._lock:
            raw_sql = self._mapper.get(mapper_name, values)
            await self.executeManyRaw(raw_sql, values)
    
    async def executeManyRaw(self, raw_sql, values=None) -> None:
        try:
            with self._lock:
                conn = await self._manager.getConnection()
                await conn.executemany(raw_sql, values)
                await self._manager.releaseConnection(conn)
        except Exception:
            if conn is not None:
                await self._manager.releaseConnection(conn)
            raise



from typing import Union, List, Any
import numpy as np
from scipy.stats import rankdata

class GroupCorrelation:
    def __init__(self, rate_min: int = 0.75, rate_max: int = 1):
        self._col_target = None
        self._rate_min = rate_min
        self._rate_max = rate_max
        self._corr_map = None
        self._corr_pos = None
        self._corr_neg = None
        self._corr_set = None
        self._corr_set_pos = None
        self._corr_set_neg = None

    def _genColumnString(self, target, idx_range: str = "pos"):
        if isinstance(target, set):
            if idx_range == "pos":
                return [self._col_target[idx] for idx in self._corr_set_pos]
            elif idx_range == "neg":
                return [self._col_target[idx] for idx in self._corr_set_neg]
            else:
                raise Exception("The vale of idx_range can only be 'pos' or 'neg'.")
        if isinstance(target, dict):
            result = {}
            for key in target.keys():
                result[self._col_target[key]] = [self._col_target[value] for value in target[key]]
            return result
        result = []
        for item in target:
            result.append([self._col_target[v] for v in item])
        return result

    def getColumnTarget(self):
        return self._col_target

    def getRelationMap(self):
        return self._corr_map

    def getRelationPositive(self, as_col_idx: bool = False):
        return self._corr_pos if as_col_idx else self._genColumnString(self._corr_pos)

    def getRelationNegative(self, as_col_idx: bool = False):
        return self._corr_neg if as_col_idx else self._genColumnString(self._corr_neg, "neg")

    def getColumnPositive(self, as_col_idx: bool = False):
        return list(self._corr_set_pos) if as_col_idx else self._genColumnString(self._corr_set_pos)

    def getColumnNegative(self, as_col_idx: bool = False):
        return list(self._corr_set_neg) if as_col_idx else self._genColumnString(self._corr_set_neg, "neg")

    def fit(self, pandas_corr, cols: List = None, method: str = "spearman", corr_abs: bool = False):
        self._col_target = pandas_corr.select_dtypes("number").columns.tolist() if cols is None else cols
        self._corr_map = pandas_corr

        if corr_abs:
            self._corr_map = np.absolute(self._corr_map)
        pos_col, pos_row = np.where((self._corr_map >= self._rate_min) & (self._corr_map <= self._rate_max))
        neg_col, neg_row = np.where((self._corr_map <= -self._rate_min) & (self._corr_map >= -self._rate_max))

        self._corr_pos = {}
        self._corr_neg = {}
        self._corr_set_pos, self._corr_set_neg = set(), set()
        for position in zip(*np.where((self._corr_map >= self._rate_min) & (self._corr_map <= self._rate_max))):
            if position[0] < position[1]:
                if position[0] in self._corr_pos:
                    self._corr_pos[position[0]].append(position[1])
                    self._corr_set_pos.add(position[1])
                else:
                    self._corr_pos[position[0]] = [position[1]]
                    self._corr_set_pos.add(position[0])
                    self._corr_set_pos.add(position[1])

        for position in zip(*np.where((self._corr_map <= -self._rate_min) & (self._corr_map >= -self._rate_max))):
            if position[0] < position[1]:
                if position[0] in self._corr_neg:
                    self._corr_neg[position[0]].append(position[1])
                    self._corr_set_neg.add(position[1])
                else:
                    self._corr_neg[position[0]] = [position[1]]
                    self._corr_set_neg.add(position[0])
                    self._corr_set_neg.add(position[1])

    def _chkGroup(self, cluster):
        check_duplicate = set.intersection(*[self.__group_dic[c] for c in cluster])
        if not check_duplicate: # if it is empty
            self.__group.append(cluster)
            for c in cluster:
                self.__group_dic[c].add(self.__group_idx)
            self.__group_idx += 1

    def _genGroup(self, data: dict, corr_set: set):
        self.__group_idx, self.__group = 0, []
        self.__group_dic = {n: set() for n in corr_set}
        for corr_key, corr_value in data.items():
            full_conn = [corr_key]
            corr_value_size = len(corr_value)
            if corr_value_size == 1:
                cluster = full_conn + corr_value
                self._chkGroup(cluster)
            else:
                for idx in range(corr_value_size):
                    sub_key = corr_value[idx]
                    sub_value = set(corr_value[idx+1:])
                    intersection = sub_value & set(data[sub_key]) if sub_key in data else set()

                    if not intersection: # 연관관계가 없을 경우 : full_conn과 sub_key를 합쳐서 append
                        cluster = full_conn + [sub_key]
                        self._chkGroup(cluster)
                    elif sub_value == intersection: # sub_set이 완전 연관관계 일 경우: full_conn에 sub_key를 append.
                        full_conn.append(sub_key)
                        if idx == corr_value_size - 1: # 만약 마지막 subset_item이면 sub_value를 append
                            full_conn.append(sub_value.pop())
                    else: # sub_set이 부분 연관관계 일 경우.
                        for inter_item in intersection:
                            cluster = full_conn + [sub_key, inter_item]
                            self._chkGroup(cluster)
                self._chkGroup(full_conn)

        return self.__group

    def getGroupPositive(self, as_col_idx: bool = False):
        corr_set = self._genGroup(self._corr_pos, self._corr_set_pos)
        return corr_set if as_col_idx else self._genColumnString(corr_set)

    def getGroupNegative(self, as_col_idx: bool = False):
        corr_set = self._genGroup(self._corr_neg, self._corr_set_neg)
        return corr_set if as_col_idx else self._genColumnString(corr_set, "neg")
    
    def getHighestGroupOfCorr(self, corr_map, cols: List = None, cnt: int = 1, positive: bool = True, negative: bool = False, abs_min: int = 0.75, abs_max: int = 1, verbose: bool = False):
        self._col_target = corr_map.columns.tolist() if cols is None else cols
        highest_group = [[col] for col in self._col_target]
        for i, tag in enumerate(self._col_target):
            tmp_map = np.concatenate([corr_map[self._col_target[i]].iloc[:i].values, np.array([0]), corr_map[self._col_target[i]].iloc[i+1:].values])
            if positive and not negative:
                ranks = rankdata(-tmp_map)
            elif not positive and negative:
                ranks = rankdata(tmp_map)
            else:
                raise Exception("Only positive or negative must be true.")
            for c in range(1, cnt+1):
                max_idx = np.where(ranks==c)[0][0]
                max_col = corr_map.columns[max_idx]
                max_val = corr_map[tag][max_col]
                if (positive and max_val<0) or (negative and max_val>0):
                    print("break")
                    break
                if abs(max_val)>=abs_min and abs(max_val)<=abs_max:
                    highest_group[i].append(max_col)
                    if verbose:
                        print(f'{tag}, {max_col} : {max_val}')
        return highest_group



import matplotlib.pyplot as plt
import math
import numpy as np
from typing import List, Union


def genSignData(freq: Union[int, List[int]], x_range: int = 1000, y_min: int = 0, y_max: int = 65536, sample: int = 1, figsize: tuple = None):
    y_range = y_max - y_min
    x_value = np.array(range(x_range * sample))
    y_value = None

    if isinstance(freq, int):
        freq = [freq]
    for f in freq:
        y_temp = np.array([(math.sin(2 * math.pi * f * x / x_range) * (y_range / 2) + ((y_max + y_min) / 2)) for x in x_value])
        y_value = y_temp if y_value is None else y_value + np.array(y_temp)
    
    min_val = np.min(y_value)
    max_val = np.max(y_value)
    y_scaled = y_min + (y_value - min_val) * (y_max - y_min) / (max_val - min_val)

    if figsize is not None:
        plt.figure(figsize=figsize)
        plt.plot(x_value, y_scaled)
        plt.title(f'Sign Wave with Frequency {freq} and Y Range {y_min} to {y_max}')
        plt.xlabel('x')
        plt.ylabel(f'Value ({y_min} to {y_max})')
        plt.grid(True)
        plt.show()
    
    return x_value, y_scaled


def genFourierInverseData(freq_sin: Union[int, List[int]], freq_cos: Union[int, List[int]], x_range: int = 1000, y_range: [int, int] = [0, 1], sample: int = 1, phase_shift: float = 0, sin_weight: float = 1, cos_weight: float = 1, figsize: tuple = None):
    """
    example : x, y = genFourierInverseData(freq_sin=2, freq_cos=2, x_range=2000, y_range=[0, 65536], sample=1, phase_shift=math.pi/4, sin_weight=1, cos_weight=0.5, figsize=(10, 4))
    """
    y_scale = y_range[1] - y_range[0]
    x_value = np.array(range(x_range * sample))
    y_value = np.zeros(len(x_value))

    if isinstance(freq_sin, int):
        freq_sin = [freq_sin]
    if isinstance(freq_cos, int):
        freq_cos = [freq_cos]

    for f in freq_sin:
        y_sin = sin_weight * np.array([(math.sin(2 * math.pi * f * x / x_range) * (y_scale / 2) + ((y_range[1] + y_range[0]) / 2)) for x in x_value])
        y_value += y_sin
    
    for f in freq_cos:
        y_cos = cos_weight * np.array([(math.cos(2 * math.pi * f * x / x_range + phase_shift) * (y_scale / 2) + ((y_range[1] + y_range[0]) / 2)) for x in x_value])
        y_value += y_cos
    
    min_val = np.min(y_value)
    max_val = np.max(y_value)
    y_scaled = y_range[0] + (y_value - min_val) * (y_range[1] - y_range[0]) / (max_val - min_val)

    if figsize is not None:
        plt.figure(figsize=figsize)
        plt.plot(x_value, y_scaled)
        plt.title(f'Fourier Series with Sin Frequencies {freq_sin} and Cos Frequencies {freq_cos}, Phase Shift {phase_shift}, and Y Range {y_range[0]} to {y_range[1]}')
        plt.xlabel('x')
        plt.ylabel(f'Value ({y_range[0]} to {y_range[1]})')
        plt.grid(True)
        plt.show()
    
    return x_value, y_scaled



from typing import Union
import pandas as pd
from .IDB import IDB

class DBPsql(IDB):
    def query(self, mapper_name: str, values: Union[dict, tuple] = None, print_sql: bool = False) -> list:
        with self._lock_q:
            raw_sql = self._mapper.get(mapper_name, values)
            if print_sql:
                print(raw_sql)
            return self.queryRaw(raw_sql, values)
    
    def queryRaw(self, raw_sql: str, values: tuple = None) -> list:
        try:
        # with self._lock_qr:
            conn = self._manager.getConnection()
            cursor = conn.cursor()

            cursor.execute(raw_sql, values)
            records = cursor.fetchall()
            
            cursor.close()
            self._manager.releaseConnection(conn)
            
            return records
        except Exception:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                self._manager.releaseConnection(conn)
            raise
    
    def queryPandas(self, mapper_name: str, values: Union[dict, tuple] = None, dtype: dict = None, print_sql: bool = False) -> pd.DataFrame:
        with self._lock_qp:
            raw_sql = self._mapper.get(mapper_name, values)
            if print_sql:
                print(raw_sql)
            return self.queryPandasRaw(raw_sql, values, dtype)
    
    def queryPandasRaw(self, raw_sql: str, values: Union[dict, tuple] = None, dtype: dict = None) -> pd.DataFrame:
        try:
        # with self._lock_qpr:
            conn = self._manager.getConnection()
            cursor = conn.cursor()

            cursor.execute(raw_sql, values)
            records = cursor.fetchall()
            columns = [desc[0] for desc in cursor.description]
            
            cursor.close()
            self._manager.releaseConnection(conn)

            df = pd.DataFrame(records, columns=columns)
            
            return df if dtype is None else df.astype(dtype)
        except Exception:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                self._manager.releaseConnection(conn)
            raise
    
    def execute(self, mapper_name, values=None, print_sql: bool = False) -> int:
        with self._lock_e:
            raw_sql = self._mapper.get(mapper_name, values)
            if print_sql:
                print(raw_sql)
            return self.executeRaw(raw_sql, values)
    
    def executeRaw(self, raw_sql, values=None) -> int:
        try:
        # with self._lock_er:
            conn = self._manager.getConnection()
            cursor = conn.cursor()

            cursor.execute(raw_sql, values)
            result = cursor.rowcount
            
            cursor.close()
            self._manager.releaseConnection(conn)
            
            return result
        except Exception:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                self._manager.releaseConnection(conn)
            raise
    
    def executeMany(self, mapper_name, values=None, print_sql: bool = False) -> int:
        with self._lock_em:
            raw_sql = self._mapper.get(mapper_name, values)
            if print_sql:
                print(raw_sql)
            return self.executeManyRaw(raw_sql, values)
    
    def executeManyRaw(self, raw_sql, values=None) -> int:
        try:
        # with self._lock_emr:
            conn = self._manager.getConnection()
            cursor = conn.cursor()

            cursor.executemany(raw_sql, values)
            result = cursor.rowcount
            
            cursor.close()
            self._manager.releaseConnection(conn)
            
            return result
        except Exception:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                self._manager.releaseConnection(conn)
            raise



import pandas as pd
import numpy as np
from typing import Union, List, Tuple, Callable

@pd.api.extensions.register_dataframe_accessor("bi")
class DanbiExtendFrame:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self._len = len(pandas_obj)

    def rollapply(self, win: int, func: Callable, test: int = 0, future: bool = False, val_nan=np.nan):
        results = []
        if test > 0:
            for idx, df in enumerate(self._obj.rolling(win)):
                if len(df) < win:
                    continue
                if idx + 1 >= win + test:
                    break
                results.append(func(df))

            return results
        else:
            for df in self._obj.rolling(win):
                if len(df) < win:
                    continue
                results.append(func(df))
            
            if len(results) > 0:
                if type(results[0]) in (list, tuple) and len(results[0]) > 1:
                    nan = [[val_nan for _ in range(len(results[0]))] for _ in range(win-1)]
                else:
                    nan = [val_nan for _ in range(win-1)]
                
            
            return results + nan if future else nan + results

    def sigmaSimilarity(self, sigma: int = 3, method: str = "sigma") -> pd.DataFrame:
        """
        method: sigma, distance
        """
        columns = self._obj.select_dtypes(include='number').columns.tolist()
        df = pd.DataFrame(1, index=columns, columns=columns)
        for idx1 in range(len(columns)):
            col1 = columns[idx1]
            min1, max1 = self._obj[col1].aot.getSigma(sigma)
            for idx2 in range(idx1+1, len(columns)):
                col2 = columns[idx2]
                min2, max2 = self._obj[col2].aot.getSigma(sigma)
                similarity = 0.0
                if method == "sigma":
                    if max1 >= min2 and max2 >= min1:
                        intersection = min(max1, max2) - max(min1, min2)
                        union = max(max1, max2) - min(min1, min2)
                        similarity = (intersection / union) * 100
                elif method == "distance":
                    amplitude1 = abs(max1 - min1)
                    amplitude2 = abs(max2 - min2)
                    similarity = min(amplitude1, amplitude2) / max(amplitude1, amplitude2) * 100
                df.iloc[idx1, idx2] = similarity
                df.iloc[idx2, idx1] = similarity
                
        return df

    def split(self, rate: float, suffle=False, seed=None) -> Tuple[pd.DataFrame, pd.DataFrame]: 
        if suffle:
            self._obj = self._obj.sample(frac=1, random_state=seed).reset_index(drop=True)
        idx = int(len(self._obj) * rate)
        return self._obj[:idx].reset_index(drop=True), self._obj[idx:].reset_index(drop=True)
    
    def timeseries(self, win: int, col_data: List[str], col_label: List[str] = None, next_label: List[int] = [1], step: int = 1) -> Tuple[np.array, np.array]:
        win_data = []
        win_label = []
        next_label = np.array(next_label)
        
        datas = self._obj[col_data].values
        if col_label is not None:
            labels = self._obj[col_label].values
        for idx in range(win, len(datas) - next_label[-1] + 1):
            if idx % step != 0:
                continue
            win_data.append(datas[idx-win:idx])
            if col_label is not None:
                win_label.append(labels[next_label + idx - 1].flatten())
        
        if col_label is None:
            return np.array(win_data), None
        else:
            return np.array(win_data), np.array(win_label)
        
    def timeseries_multi(self, win: int, col_datas: List[List[str]], col_labels: List[List[str]] = None, next_label: List[int] = [1], step: int = 1) -> Tuple[np.array, np.array]:
        win_data = [[] for _ in range(len(col_datas))]
        win_label = [[] for _ in range(len(col_labels))]
        next_label = np.array(next_label)
 
        datas = []
        labels = []
        for col_data in col_datas:
            datas.append(self._obj[col_data].values)
        if col_labels is not None:
            for col_label in col_labels:
                labels.append(self._obj[col_label].values)
        for idx in range(win, len(datas[0]) - next_label[-1] + 1):
            if idx % step != 0:
                continue
            for idx_array in range(len(col_datas)):
                win_data[idx_array].append(datas[idx_array][idx-win:idx])
            if col_label is not None:
                for idx_array in range(len(col_labels)):
                    win_label[idx_array].append(labels[idx_array][next_label + idx - 1].flatten())
        if col_label is None:
            return np.array(win_data), None
        else:
            return np.array(win_data), np.array(win_label)

    def dtype(self, src: Union[str, List[str]], dst: str, inplace: bool = True) -> pd.DataFrame:
        """Change all columns of a specific type to the target type. support numpy like notation.
    
        Args:
            src (List[str]): list of types to change.
                > optional : you can use the numpy like type notation.
                > example: i : int, int16, int32 ...
                           f : float, float16, float32 ...
            dst (str): target type.
            inplace (bool): apply the changed data directly to the original data.

        Examples:
           > df.bi.dtype(["i", "f"], "int8")
           > df.bi.dtype("f", "int8", True)
           > df_changed = df.bi.dtype("float8", "int8", False)
    
        Returns:
            When the inplace option is set to False, the method returns a pandas DataFrame that has been changed.
        """
        if isinstance(src, str):
            src = [src]
        
        cols = []
        for item in src:
            if len(item) == 1:
                cols += self._obj.columns[[dtype.kind in src for dtype in self._obj.dtypes]].tolist()
            else:
                cols += self._obj.columns[self._obj.dtypes.isin([np.dtype(ty) for ty in src])].tolist()
        cols = set(cols)
        
        if inplace:
            for col in cols:
                self._obj[col] = self._obj[col].astype(dst)
        else:
            return self._obj.astype({col: dst for col in cols})


import abc

class IPluginAsync(abc.ABC):
    def __new__(self, name):
        if not hasattr(self, 'instance'):
            self.instance = super(IPluginAsync, self).__new__(self)
            self._name = name
        return self.instance
    
    def getName(self):
        return self._name
    
    @abc.abstractmethod
    async def plug(self, **kwargs) -> bool:
        ...
    
    @abc.abstractmethod
    async def unplug(self, **kwargs) -> bool:
        ...
    
    def __repr__(self):
        return self._name



import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression

model_linear = LinearRegression()
def anaRegressionTrend(data, outlier_rate=0.1, span=0):
    global model_linear

    # 선형회귀 계산
    size_data = len(data)
    time = np.arange(size_data).reshape(-1, 1)
    model_linear.fit(time, data)
    slope = model_linear.coef_[0] # 기울기
    intercept = model_linear.intercept_ # 절편
    trend = model_linear.predict(np.arange(size_data+span).reshape(-1, 1))

    # 상한 & 하한 선 계산
    series_diff = pd.Series(data - trend[:size_data])
    lower = series_diff[series_diff <= series_diff.quantile(outlier_rate)].abs().mean()
    upper = series_diff[series_diff >= series_diff.quantile(1- outlier_rate)].abs().mean()
    median = upper if upper < lower else lower

    upper_intercept = intercept + median
    lower_intercept = intercept - median
    
    # 상위 10%, 하위 10% 라인 계산
    trend_upper = (slope * np.arange(size_data+span).reshape(-1, 1) + upper_intercept).flatten()
    trend_lower = (slope * np.arange(size_data+span).reshape(-1, 1) + lower_intercept).flatten()

    return slope, intercept, trend, trend_upper, trend_lower


from .ZeroBaseMinMaxScaler import ZeroBaseMinMaxScaler, getMinMaxRows
from .TimeseriesToImage import genFigureImageArray, ITimeEncoder, TimeEncodingBuilder
from .GroupCorrelation import GroupCorrelation
from .DataBalance import UniqueRateChecker, LabelRateAugmenter
from .convert import convDfsToContinuousDfs
from .signal import genSignData, genFourierInverseData
from .analysis import anaRegressionTrend


import pandas as pd
from typing import List

def convDfsToContinuousDfs(dfs: List[pd.DataFrame], time_column: str, freq: str, replace_method: str = "mean") -> List[pd.DataFrame]:
    '''
    example : res = convDfsToContinuousDfs([df1, df2], "Timestamp", "1d", "min")
    '''
    df = pd.concat(dfs).drop_duplicates().sort_values(by=time_column).reset_index(drop=True)    
    df.set_index(time_column, inplace=True)
    if replace_method == "mean":
        df = df.resample(freq).mean(numeric_only=True).dropna()
    elif replace_method == "median":
        df = df.resample(freq).median(numeric_only=True).dropna()
    elif replace_method == "min":
        df = df.resample(freq).min(numeric_only=True).dropna()
    elif replace_method == "max":
        df = df.resample(freq).max(numeric_only=True).dropna()
    else:
        raise Exception('Please enter one of four values: mean, median, min_max_mean.')
    df.reset_index(inplace=True)
    
    delta = pd.Timedelta(freq)
    offsets = df[[time_column]].diff()
    offsets.iloc[0, 0] = delta
    points = offsets[offsets[time_column] != delta].index
    offsets = offsets.drop(time_column, axis=1)
    
    point_start = 0
    df_results = []
    for point in points:
        df_results.append(offsets[point_start:point].merge(df, left_index=True, right_index=True))
        point_start = point
    df_results.append(offsets[point_start:].merge(df, left_index=True, right_index=True))

    return df_results



import abc
import cv2
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Callable


def genFigureImageArray(fig: plt.figure, width: int = None, height: int = None, channels: str = "r"):
    fig.canvas.draw()
    fig_data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    image = fig_data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    if width is not None or height is not None:
        image = cv2.resize(image, (width, height))

    imgarrays = []
    imgarray = np.array(image)
    for channel in channels:
        imgarrays.append(imgarray[:, :, "rgb".index(channel)])
    imgarray = np.max(imgarrays, axis=0)

    return imgarray * (255 // imgarray.max())


class ITimeEncoder(abc.ABC):
    def __init__(self, slice_start: int = None, slice_end: int = None):
        self._slice_start = slice_start
        self._slice_end = slice_end
    
    @abc.abstractclassmethod
    def setParams(self, **kwargs):
        ...
    
    @abc.abstractclassmethod
    def getEncoded(self, data: np.array) -> List[np.array]:
        ...


class TimeEncodingBuilder():
    def __init__(self, encoders: List[ITimeEncoder]):
        self._encoders = encoders
    
    def getImage(self, data: np.array, tolist: bool = False):
        encoded = []
        for encoder in self._encoders:
            encoded += encoder.getEncoded(data)
        
        return encoded if tolist else np.stack(encoded, axis=-1)
    
    def getTimeseries(self, data: np.array, tolist: bool = False):
        encoded = []
        for encoder in self._encoders:
            encoded += encoder.getEncoded(data)

        return encoded if tolist else np.concatenate(encoded)




import sys, re
import pkgutil

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata

def infoInstalledPackage(name: str = None, version: str = None, license: str = None) -> list:
    """ Find information about all packages installed on the system.

    Args:
        name (str, optional): regular expression of packages name. Defaults to None.
        version (str, optional): regular expression of packages version. Defaults to None.
        license (str, optional): regular expression of packages license. Defaults to None.

    Returns:
        _type_: list of tuple. like '[(name, version, license)]'
    """
    regex_name = re.compile(name, re.I) if name else None
    regex_version = re.compile(version, re.I) if version else None
    regex_license = re.compile(license, re.I) if license else None
    
    dists = importlib_metadata.distributions()
    results = []
    for dist in dists:
        name = dist.metadata["Name"]
        version = dist.version
        license = dist.metadata["License"]
        
        if regex_license and not license: continue
        if regex_name and not bool(regex_name.search(name)): continue
        if regex_version and not bool(regex_version.search(version)): continue
        if regex_license and not bool(regex_license.search(license)): continue
        
        results.append((name, version, license))
    
    return results

def infoSubmodules(base: str = None, package: bool = True, include_path: bool = False) -> list:
    """Find all sub-module name or sub-package name.

    Args:
        base (str, optional): base name of package. Defaults to None.
        package (bool, optional): true to find package name, false to find file name. Defaults to True.
        include_path (bool, optional): true if you want to append the path name to the result.. Defaults to False.

    Returns:
        list: list of sub-module or sub-package name
    """
    pkg_base = __import__(base, fromlist=["blah"])
    results = []
    for loader, module_name, is_pkg in pkgutil.walk_packages(pkg_base.__path__, pkg_base.__name__ + "."):
        if is_pkg is package:
            results.append([loader.path, module_name] if include_path else module_name)
    return results


from .info import (
    infoInstalledPackage,
    infoSubmodules
)
from .day_time import (
    MO, TU, WE, TH, FR, SA, SU,
    getDaysBetween,
    getDayPeriod,
    getWeekday, getYearAnimal, getMonthStartEnd
)
from .jupyter import (
    asClassMethod, setWarningOff, showAsCols
)
from .util import (
    storePickle, restorePickle
)
from .with_print import WithNumpyPrint, WithPandasPrint
from .EduPlot import EduPlotConf, EduPlot2D
from .DotDict import DotDict


import numpy as np
import pandas as pd
from typing import List

def getStateUpDn(df: pd.DataFrame, window: int, col: str, rate_dn: float = 1, rate_up: float = 1.1) -> List:
    ups = df[col] == df[col].rolling(window).min().shift(-window+1)
    dns = df[col] == df[col].rolling(window).max().shift(-window+1)
    updn, term, is_up, is_change, s_val, e_val = [], [], False, False, 0, 0
    
    for up, dn, val in zip(ups, dns, df[col]):
        if not np.isnan(val):
            e_val = val
        if is_change:
            updn += term
            term, is_change = [], False
            s_val = val
        term.append(is_up)
        
        if not is_up and up: # change state: dn -> up
            is_up, is_change = True, True
            if s_val != 0 and s_val * rate_dn < val: # dn state does not satisfy rate_dn condition.
                term = list(np.logical_not(term))
        elif is_up and dn: # change state: up -> dn
            is_up, is_change = False, True
            if s_val != 0 and s_val * rate_up > val: # up state does not satisfy rate_up condition.
                term = list(np.logical_not(term))
    
    if (is_up and s_val * rate_up > e_val) or (not is_up and s_val * rate_dn < e_val):
        term = list(np.logical_not(term))
    updn += term

    return updn


import abc

class IConnectionManager(abc.ABC):
    def __new__(self, **kwargs):
        self._kwargs = kwargs
        if not hasattr(self, 'instance'):
            self.instance = super(IConnectionManager, self).__new__(self)
            self._conn_pool = None
        return self.instance
    
    @abc.abstractclassmethod
    def connect(self, **kwargs):
        ...
    
    @abc.abstractclassmethod
    def isConnect(self) -> bool:
        ...

    @abc.abstractclassmethod
    def close(self, **kwargs):
        ...
    
    @abc.abstractclassmethod
    def getConnection(self, auto_commit=True, **kwargs):
        ...
    
    @abc.abstractclassmethod
    def releaseConnection(self, conn):
        ...
